@@ -1,6 +1,7 @@
 import warnings
 from typing import Sequence
 
+from numpy import concatenate
 from numpy import isscalar
 from numpy import ndim
 from numpy import newaxis
@@ -20,6 +21,7 @@ from pysindy.optimizers import SINDyOptimizer
 from pysindy.optimizers import STLSQ
 from pysindy.utils.base import drop_nan_rows
 from pysindy.utils.base import equations
+from pysindy.utils.base import validate_control_variables
 from pysindy.utils.base import validate_input
 
 
@@ -31,25 +33,29 @@ class SINDy(BaseEstimator):
     ----------
     optimizer : optimizer object, optional
         Optimization method used to fit the SINDy model. This must be an object
-        that extends the sindy.optimizers.BaseOptimizer class. Default is
+        extending the ``sindy.optimizers.BaseOptimizer`` class. Default is
         sequentially thresholded least squares with a threshold of 0.1.
 
     feature_library : feature library object, optional
+        Feature library object used to specify candidate right-hand side features.
+        This must be an object extending the
+        ``sindy.feature_library.BaseFeatureLibrary`` class.
         Default is polynomial features of degree 2.
 
     differentiation_method : differentiation object, optional
-        Method for differentiating the data. This must be an object that
-        extends the sindy.differentiation_methods.BaseDifferentiation class.
+        Method for differentiating the data. This must be an object extending
+        the ``sindy.differentiation_methods.BaseDifferentiation`` class.
         Default is centered difference.
 
     feature_names : list of string, length n_input_features, optional
-        Names for the input features. If None, will use ['x0','x1',...].
+        Names for the input features (e.g. ``['x', 'y', 'z']``). If None, will use
+        ``['x0', 'x1', ...]``.
 
     discrete_time : boolean, optional (default False)
         If True, dynamical system is treated as a map. Rather than predicting
         derivatives, the right hand side functions step the system forward by
         one time step. If False, dynamical system is assumed to be a flow
-        (right hand side functions predict continuous time derivatives).
+        (right-hand side functions predict continuous time derivatives).
 
     n_jobs : int, optional (default 1)
         The number of parallel jobs to use when fitting, predicting with, and
@@ -60,6 +66,13 @@ class SINDy(BaseEstimator):
     model : sklearn.multioutput.MultiOutputRegressor object
         The fitted SINDy model.
 
+    n_input_features_ : int
+        The total number of input features.
+
+    n_output_features_ : int
+        The total number of output features. This number is a function of
+        ``self.n_input_features`` and the feature library being used.
+
     Examples
     --------
     >>> import numpy as np
@@ -89,26 +102,70 @@ class SINDy(BaseEstimator):
            [ 0.        ,  0.        ,  0.        ]])
     >>> model.score(x, t=t[1]-t[0])
     0.999999985520653
+
+    >>> import numpy as np
+    >>> from scipy.integrate import odeint
+    >>> from pysindy import SINDy
+    >>> u = lambda t : np.sin(2 * t)
+    >>> lorenz_c = lambda z,t : [
+                10 * (z[1] - z[0]) + u(t) ** 2,
+                z[0] * (28 - z[2]) - z[1],
+                z[0] * z[1] - 8 / 3 * z[2],
+        ]
+    >>> t = np.arange(0,2,0.002)
+    >>> x = odeint(lorenz_c, [-8,8,27], t)
+    >>> u_eval = u(t)
+    >>> model = SINDy()
+    >>> model.fit(x, u_eval, t=t[1]-t[0])
+    >>> model.print()
+    x0' = -10.000 x0 + 10.000 x1 + 1.001 u0^2
+    x1' = 27.994 x0 + -0.999 x1 + -1.000 x0 x2
+    x2' = -2.666 x2 + 1.000 x0 x1
+    >>> model.coefficients()
+    array([[ 0.        , -9.99969851,  9.99958359,  0.        ,  0.        ,
+             0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
+             0.        ,  0.        ,  0.        ,  0.        ,  1.00120331],
+           [ 0.        , 27.9935177 , -0.99906375,  0.        ,  0.        ,
+             0.        ,  0.        , -0.99980455,  0.        ,  0.        ,
+             0.        ,  0.        ,  0.        ,  0.        ,  0.        ],
+           [ 0.        ,  0.        ,  0.        , -2.666437  ,  0.        ,
+             0.        ,  0.99990137,  0.        ,  0.        ,  0.        ,
+             0.        ,  0.        ,  0.        ,  0.        ,  0.        ]])
+    >>> model.score(x, u_eval, t=t[1]-t[0])
+    0.9999999855414495
     """
 
     def __init__(
         self,
-        optimizer=STLSQ(),
-        feature_library=PolynomialFeatures(),
-        differentiation_method=FiniteDifference(),
+        optimizer=None,
+        feature_library=None,
+        differentiation_method=None,
         feature_names=None,
         discrete_time=False,
         n_jobs=1,
     ):
+        if optimizer is None:
+            optimizer = STLSQ()
         self.optimizer = optimizer
+        if feature_library is None:
+            feature_library = PolynomialFeatures()
         self.feature_library = feature_library
+        if differentiation_method is None:
+            differentiation_method = FiniteDifference()
         self.differentiation_method = differentiation_method
         self.feature_names = feature_names
         self.discrete_time = discrete_time
         self.n_jobs = n_jobs
 
     def fit(
-        self, x, t=1, x_dot=None, multiple_trajectories=False, unbias=True, quiet=False
+        self,
+        x,
+        t=1,
+        x_dot=None,
+        u=None,
+        multiple_trajectories=False,
+        unbias=True,
+        quiet=False,
     ):
         """
         Fit the SINDy model.
@@ -139,6 +196,15 @@ class SINDy(BaseEstimator):
             provided, it must match the shape of the training data and these
             values will be used as the time derivatives.
 
+        u: array-like or list of array-like, shape (n_samples, n_control_features), \
+                optional (default None)
+            Control variables/inputs. Include this variable to use sparse
+            identification for nonlinear dynamical systems for control (SINDYc).
+            If training data contains multiple trajectories (i.e. if x is a list of
+            array-like), then u should be a list containing control variable data
+            for each trajectory. Individual trajectories may contain different
+            numbers of samples.
+
         multiple_trajectories: boolean, optional, (default False)
             Whether or not the training data includes multiple trajectories. If
             True, the training data must be a list of arrays containing data
@@ -148,10 +214,10 @@ class SINDy(BaseEstimator):
         unbias: boolean, optional (default True)
             Whether to perform an extra step of unregularized linear regression to
             unbias the coefficients for the identified support.
-            If the optimizer (`SINDy.optimizer`) applies any type of regularization,
+            If the optimizer (``SINDy.optimizer``) applies any type of regularization,
             that regularization may bias coefficients toward particular values,
             improving the conditioning of the problem but harming the quality of the
-            fit. Setting `unbias=True` enables an extra step wherein unregularized
+            fit. Setting ``unbias=True`` enables an extra step wherein unregularized
             linear regression is applied, but only for the coefficients in the support
             identified by the optimizer. This helps to remove the bias introduced by
             regularization.
@@ -163,6 +229,18 @@ class SINDy(BaseEstimator):
         -------
         self: returns an instance of self
         """
+        if u is None:
+            self.n_control_features_ = 0
+        else:
+            trim_last_point = self.discrete_time and (x_dot is None)
+            u = validate_control_variables(
+                x,
+                u,
+                multiple_trajectories=multiple_trajectories,
+                trim_last_point=trim_last_point,
+            )
+            self.n_control_features_ = u.shape[1]
+
         if multiple_trajectories:
             x, x_dot = self.process_multiple_trajectories(x, t, x_dot)
         else:
@@ -173,13 +251,17 @@ class SINDy(BaseEstimator):
                     x_dot = x[1:]
                     x = x[:-1]
                 else:
-                    x_dot = validate_input(x)
+                    x_dot = validate_input(x_dot)
             else:
                 if x_dot is None:
                     x_dot = self.differentiation_method(x, t)
                 else:
                     x_dot = validate_input(x_dot, t)
 
+        # Append control variables
+        if self.n_control_features_ > 0:
+            x = concatenate((x, u), axis=1)
+
         # Drop rows where derivative isn't known
         x, x_dot = drop_nan_rows(x, x_dot)
 
@@ -200,13 +282,15 @@ class SINDy(BaseEstimator):
 
         if self.feature_names is None:
             feature_names = []
-            for i in range(self.n_input_features_):
+            for i in range(self.n_input_features_ - self.n_control_features_):
                 feature_names.append("x" + str(i))
+            for i in range(self.n_control_features_):
+                feature_names.append("u" + str(i))
             self.feature_names = feature_names
 
         return self
 
-    def predict(self, x, multiple_trajectories=False):
+    def predict(self, x, u=None, multiple_trajectories=False):
         """
         Predict the time derivatives using the SINDy model.
 
@@ -215,6 +299,12 @@ class SINDy(BaseEstimator):
         x: array-like or list of array-like, shape (n_samples, n_input_features)
             Samples.
 
+        u: array-like or list of array-like, shape(n_samples, n_control_features), \
+                (default None)
+            Control variables. If ``multiple_trajectories=True`` then u
+            must be a list of control variable data from each trajectory. If the
+            model was fit with control variables then u is not optional.
+
         multiple_trajectories: boolean, optional (default False)
             If True, x contains multiple trajectories and must be a list of
             data from each trajectory. If False, x is a single trajectory.
@@ -225,12 +315,36 @@ class SINDy(BaseEstimator):
             Predicted time derivatives
         """
         check_is_fitted(self, "model")
-        if multiple_trajectories:
-            x = [validate_input(xi) for xi in x]
-            return [self.model.predict(xi) for xi in x]
+        if u is None or self.n_control_features_ == 0:
+            if self.n_control_features_ > 0:
+                raise TypeError(
+                    "Model was fit using control variables, so u is required"
+                )
+            elif u is not None:
+                warnings.warn(
+                    "Control variables u were ignored because control variables were"
+                    " not used when the model was fit"
+                )
+            if multiple_trajectories:
+                x = [validate_input(xi) for xi in x]
+                return [self.model.predict(xi) for xi in x]
+            else:
+                x = validate_input(x)
+                return self.model.predict(x)
         else:
-            x = validate_input(x)
-            return self.model.predict(x)
+            if multiple_trajectories:
+                x = [validate_input(xi) for xi in x]
+                u = validate_control_variables(
+                    x, u, multiple_trajectories=True, return_array=False
+                )
+                return [
+                    self.model.predict(concatenate((xi, ui), axis=1))
+                    for xi, ui in zip(x, u)
+                ]
+            else:
+                x = validate_input(x)
+                u = validate_control_variables(x, u)
+                return self.model.predict(concatenate((x, u), axis=1))
 
     def equations(self, precision=3):
         """
@@ -281,6 +395,7 @@ class SINDy(BaseEstimator):
         x,
         t=1,
         x_dot=None,
+        u=None,
         multiple_trajectories=False,
         metric=r2_score,
         **metric_kws
@@ -300,12 +415,18 @@ class SINDy(BaseEstimator):
             provided.
 
         x_dot: array-like or list of array-like, shape (n_samples, n_input_features), \
-                optional
+                optional (default None)
             Optional pre-computed derivatives of the samples. If provided,
             these values will be used to compute the score. If not provided,
             the time derivatives of the training data will be computed using
             the specified differentiation method.
 
+        u: array-like or list of array-like, shape(n_samples, n_control_features), \
+                optional (default None)
+            Control variables. If ``multiple_trajectories=True`` then u
+            must be a list of control variable data from each trajectory.
+            If the model was fit with control variables then u is not optional.
+
         multiple_trajectories: boolean, optional (default False)
             If True, x contains multiple trajectories and must be a list of
             data from each trajectory. If False, x is a single trajectory.
@@ -322,6 +443,25 @@ class SINDy(BaseEstimator):
         score: float
             Metric function value for the model prediction of x_dot.
         """
+        if u is None or self.n_control_features_ == 0:
+            if self.n_control_features_ > 0:
+                raise TypeError(
+                    "Model was fit using control variables, so u is required"
+                )
+            elif u is not None:
+                warnings.warn(
+                    "Control variables u were ignored because control variables were"
+                    " not used when the model was fit"
+                )
+        else:
+            trim_last_point = self.discrete_time and (x_dot is None)
+            u = validate_control_variables(
+                x,
+                u,
+                multiple_trajectories=multiple_trajectories,
+                trim_last_point=trim_last_point,
+            )
+
         if multiple_trajectories:
             x, x_dot = self.process_multiple_trajectories(
                 x, t, x_dot, return_array=True
@@ -338,6 +478,10 @@ class SINDy(BaseEstimator):
         if ndim(x_dot) == 1:
             x_dot = x_dot.reshape(-1, 1)
 
+        # Append control variables
+        if u is not None and self.n_control_features_ > 0:
+            x = concatenate((x, u), axis=1)
+
         # Drop rows where derivative isn't known (usually endpoints)
         x, x_dot = drop_nan_rows(x, x_dot)
 
@@ -353,12 +497,10 @@ class SINDy(BaseEstimator):
             raise TypeError("Input x must be a list")
 
         if self.discrete_time:
+            x = [validate_input(xi) for xi in x]
             if x_dot is None:
-                x_dot = []
-                for i in range(len(x)):
-                    x_tmp = validate_input(x[i])
-                    x[i] = x_tmp[:-1]
-                    x_dot.append(x_tmp[1:])
+                x_dot = [xi[1:] for xi in x]
+                x = [xi[:-1] for xi in x]
             else:
                 if not isinstance(x_dot, Sequence):
                     raise TypeError(
@@ -369,15 +511,13 @@ class SINDy(BaseEstimator):
         else:
             if x_dot is None:
                 if isinstance(t, Sequence):
-                    x_dot = []
-                    for i in range(len(x)):
-                        x[i] = validate_input(x[i], t[i])
-                        x_dot.append(self.differentiation_method(x[i], t[i]))
+                    x = [validate_input(xi, ti) for xi, ti in zip(x, t)]
+                    x_dot = [
+                        self.differentiation_method(xi, ti) for xi, ti in zip(x, t)
+                    ]
                 else:
-                    x_dot = []
-                    for i in range(len(x)):
-                        x[i] = validate_input(x[i], t)
-                        x_dot.append(self.differentiation_method(x[i], t))
+                    x = [validate_input(xi, t) for xi in x]
+                    x_dot = [self.differentiation_method(xi, t) for xi in x]
             else:
                 if not isinstance(x_dot, Sequence):
                     raise TypeError(
@@ -385,8 +525,10 @@ class SINDy(BaseEstimator):
                         "(i.e. for multiple trajectories)"
                     )
                 if isinstance(t, Sequence):
-                    x_dot = [validate_input(xd, t) for xd, t in zip(x_dot, t)]
+                    x = [validate_input(xi, ti) for xi, ti in zip(x, t)]
+                    x_dot = [validate_input(xd, ti) for xd, ti in zip(x_dot, t)]
                 else:
+                    x = [validate_input(xi, t) for xi in x]
                     x_dot = [validate_input(xd, t) for xd in x_dot]
 
         if return_array:
@@ -441,7 +583,9 @@ class SINDy(BaseEstimator):
             input_features=self.feature_names
         )
 
-    def simulate(self, x0, t, integrator=odeint, stop_condition=None, **integrator_kws):
+    def simulate(
+        self, x0, t, u=None, integrator=odeint, stop_condition=None, **integrator_kws
+    ):
         """
         Simulate the SINDy model forward in time.
 
@@ -455,6 +599,16 @@ class SINDy(BaseEstimator):
             points at which to simulate. If the model is in discrete time,
             t must be an integer indicating how many steps to predict.
 
+        u: function from R^1 to R^{n_control_features} or list/array, optional \
+            (default None)
+            Control input function.
+            If the model is continuous time, i.e. ``self.discrete_time == False``,
+            this function should take in a time and output the values of each of
+            the n_control_features control features as a list or numpy array.
+            If the model is discrete time, i.e. ``self.discrete_time == True``,
+            u should be a list (with ``len(u) == t``) or array (with
+            ``u.shape[0] == 1``) giving the control inputs at each step.
+
         integrator: function object, optional
             Function to use to integrate the system. Default is scipy's odeint.
 
@@ -470,6 +624,9 @@ class SINDy(BaseEstimator):
         x: numpy array, shape (n_samples, n_features)
             Simulation results
         """
+        check_is_fitted(self, "model")
+        if u is None and self.n_control_features_ > 0:
+            raise TypeError("Model was fit using control variables, so u is required")
 
         if self.discrete_time:
             if not isinstance(t, int):
@@ -478,12 +635,34 @@ class SINDy(BaseEstimator):
                     "the number of steps to predict)"
                 )
 
-            x = zeros((t, self.n_input_features_))
+            if stop_condition is not None:
+
+                def check_stop_condition(xi):
+                    return stop_condition(xi)
+
+            else:
+
+                def check_stop_condition(xi):
+                    pass
+
+            x = zeros((t, self.n_input_features_ - self.n_control_features_))
             x[0] = x0
-            for i in range(1, t):
-                x[i] = self.predict(x[i - 1 : i])
-                if stop_condition is not None and stop_condition(x[i]):
-                    return x[: i + 1]
+
+            if u is None or self.n_control_features_ == 0:
+                if u is not None:
+                    warnings.warn(
+                        "Control variables u were ignored because control variables"
+                        " were not used when the model was fit"
+                    )
+                for i in range(1, t):
+                    x[i] = self.predict(x[i - 1 : i])
+                    if check_stop_condition(x[i]):
+                        return x[: i + 1]
+            else:
+                for i in range(1, t):
+                    x[i] = self.predict(x[i - 1 : i], u=u[i - 1])
+                    if check_stop_condition(x[i]):
+                        return x[: i + 1]
             return x
         else:
             if isscalar(t):
@@ -492,8 +671,20 @@ class SINDy(BaseEstimator):
                     " points at which to simulate"
                 )
 
-            def rhs(x, t):
-                return self.predict(x[newaxis, :])[0]
+            if u is None or self.n_control_features_ == 0:
+                if u is not None:
+                    warnings.warn(
+                        "Control variables u were ignored because control variables"
+                        " were not used when the model was fit"
+                    )
+
+                def rhs(x, t):
+                    return self.predict(x[newaxis, :])[0]
+
+            else:
+
+                def rhs(x, t):
+                    return self.predict(x[newaxis, :], u(t))[0]
 
             return integrator(rhs, x0, t, **integrator_kws)
 
